/***************************************************************************
                          ssl.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2009  Marian VooDooMan Meravy
    email                : ghostvoodooman@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2009  Marian VooDooMan Meravy
  (ghostvoodooman@users.sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides interface to SSL library
*/

#ifndef SSL_H__
#define SSL_H__

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <deque>

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

//#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"

#ifdef USE_OPENSSL
#   include <openssl/rsa.h>       /* SSLeay stuff */
#   include <openssl/crypto.h>
#   include <openssl/x509.h>
#   include <openssl/pem.h>
#   include <openssl/ssl.h>
#   include <openssl/err.h>
#   include <openssl/x509v3.h>
#endif

#include "compat.h"

#ifndef USE_OPENSSL
#   include <gnutls/gnutls.h>
#endif

#include "sock.h"
#include "sock_shared.h"

#ifdef USE_OPENSSL
typedef SSL* ssl_handle;
#else
typedef gnutls_session ssl_handle;
#endif

#include "assert.h"

using namespace std;

/*!
    \brief Stores configuration data (ssl.txt) for one bot to accept its connection via SSL
    \author VooDooMan
    \version 1
    \date 2004
*/
typedef struct {
    string botname;                 //!< Name of bot from logic.txt
    string cert;                    //!< Path to its client's certificate
    string key;                        //!< Path to client's key (when connecting to IRC server)
} s_ssl_bot;

/*!
    \brief Stores configuration from ssl.txt
    \author VooDooMan
    \version 1
    \date 2004
*/
typedef struct {
    string accept_ca_certs;         //!< File containing certificates of CA's that we are trust
    string server_cert;             //!< Path to file with server's certificate chain
    string server_key;              //!< Path to file with server's private key
    string client_cert;             //!< Path to file with client certificate chain we use to connect as client to another bot
    string client_key;              //!< Path to file with key for client certificate
    vector<s_ssl_bot> ssl_bots;     //!< Array of bots to use SSL for incomming connection
} s_ssl_conf;

string ssl_get_ssl_version();

void ssl_loop();

/*!
    \brief Stores data to send/receive later
    \author VooDooMan
    \version 2
    \date 2004, 2006-07-11
*/
struct s_ssl_data {
    size_t len;                         //!< Length of data
    void* data;                         //!< Data
};

/*!
    \brief SSL descriptor
    \author VooDooMan
    \version 1
    \date 2006-07-11
*/
class c_ssl {
    s_socket* s;                                //!< Socket
    ssl_handle handle;                          //!< SSL handle

#ifndef USE_OPENSSL
    friend ssize_t pull_func(gnutls_transport_ptr handle, void* data, size_t size);
    friend ssize_t push_func(gnutls_transport_ptr handle, const void* data, size_t size);
#endif

#ifdef USE_OPENSSL
    BIO* rbio;                                  //!< BIO structure for OpenSSL for reading
    BIO* wbio;                                  //!< BIO structure for OpenSSL for writing
#endif

    bool cert_ok_so_far;                        //!< If certificate was verified okay so far
    string expected_cert_file;                  //!< Expected certificate (filename)

    deque<s_ssl_data> to_read;                  //!< Chunks to read
    deque<s_ssl_data> to_write;                 //!< Chunks to write

    bool error;                                 //!< If there was error
    string err_str;                             //!< Error string

    enum e_state {                              //!< State code
        disconnected = 0,                       //!< Disconnected
        want_handshake,                         //!< Waiting for handshake (SSL_connect)
        want_accept,                            //!< Waiting for handshake (SSL_accept)
        verify_cert,                            //!< Do a certificate verify
        session_ok,                             //!< Okay to communicate, certificate valid, etc.
        in_shutdown,                            //!< Shutdown state
    };

    e_state state;                              //!< State
    bool force_write;                           //!< Should be called write function again first

    /*!
        \brief Tries to do handshake on the client connection
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    void do_handshake();

    /*!
        \brief Tries to do handshake on the server connection
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    void do_accept();

    /*!
        \brief Does transfer
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    void do_transfer();

    /*!
        \brief Verifies the peer's certificate
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    void do_verify();

    /*!
        \brief Does a shutdown
        \author VooDooMan
        \version 1
        \date 2006-07-12
    */
    void do_shutdown();

    /*!
        \brief Gets OpenSSL error
        \author VooDooMan
        \version 1
        \date 2006-07-12
    */
    static string get_openssl_error();

    /*!
        \brief Gets OpenSSL error from result code
        \author VooDooMan
        \version 1
        \date 2006-07-12
        \param err Return code of last operation
    */
    string get_openssl_error(unsigned long err);

    /*!
        \brief Clears error queue
        \author VooDooMan
        \version 1
        \date 2006-07-12
    */
    static void clear_error_queue();

    /*!
        \brief Searches for, and destroys meta any data
        \author VooDooMan
        \version 1
        \date 2007-02-27
    */
    void destroy();

public:

#ifdef USE_OPENSSL
    static int my_verify_callback(int preverify_ok, X509_STORE_CTX *ctx);
#endif

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param socket Socket
        \param server Server socket, or NULL for client
        \param cert Filename with certificate
        \param key Filename with key
    */
    c_ssl(s_socket& socket, s_socket* server, const string& cert, const string& key);

    /*!
        \brief Reads from SSL
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param buff Buffer
        \param len Length of buffer
        \return Returns number of read bytes
    */
    size_t read(void* buff, size_t len);

    /*!
        \brief Writes to SSL
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param buff Buffer
        \param len Length of buffer
    */
    void write(const void* buff, size_t len);

    /*!
        \brief Gets error
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param str Returns error string
        \return Returns true for error
    */
    bool get_error(string& str);

    /*!
        \brief Set certificate to expect
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param cert_file Filename with certificate
    */
    void set_expected_cert_file(const string& cert_file);

    /*!
        \brief Retrieves statistics about sock I/O cache
        \author VooDooMan
        \version 2
        \date 2006-01-22, 2006-07-11
        \param to_receive Returns number of bytes waiting in cache to receive them
        \param to_send Returns number of bytes waiting in cache to send them
    */
    void get_cache_stats(size_t& to_receive, size_t& to_send);

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    ~c_ssl();

    /*!
        \brief Processes SSL I/O transactions
        \author VooDooMan
        \version 1
        \date 2006-07-11
    */
    void process();

    /*!
        \brief If there was no error so far, set one
        \author VooDooMan
        \version 1
        \date 2006-07-11
        \param err_string Error string to set
    */
    void set_error(const string& err_string);

    /*!
        \brief Initiate shutdown
        \author VooDooMan
        \version 1
        \date 2006-07-12
    */
    void shutdown();

    /*!
        \brief Returns true if we are connected
        \author VooDooMan
        \version 1
        \date 2006-07-12
    */
    bool is_open();
};

bool ssl_init(string& err_str, char* trustedCAs);
void ssl_done();
void ssl_put_to_garbage(c_ssl* ssl);

#endif

